#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import re
from inspect import signature
from io import StringIO
from typing import Dict, Union, Callable

from garoupa import identity32, identity64, Hosh
from uncompyle6.main import decompile

from ldict.abs.mixin.aux import Aux, VT
from ldict.data import process, fhosh


# Dict typing inheritance initially based on https://stackoverflow.com/a/64323140/9681577
# TODO: aceitar qq class que sirva no lugar de Hosh, em vez da string 'version'. Requer Hosh32 e Hosh64 herdando de Hosh
#           serve pra viabilizar teste exaustivo de grupo pequeno pra ver se o artigo funciona
class Ldict_(Aux, Dict[str, VT]):
    """Uniquely identified lazy dict for serializable pairs str->value
    (serializable in the project 'orjson' sense).

    Usage:

    >>> from ldict import ldict
    >>> print(ldict(x=123123, y=88))
    {
        "id": "0000000000000000000004fXFwHzGuTlkQcR5z1rzEXbWAnrZUJOq5Ua3YnFcvvx",
        "ids": "<1 hidden ids>",
        "x": 123123,
        "y": 88
    }
    >>> print(ldict(y=88, x=123123))  # Original values are order-insensitive.
    {
        "id": "0000000000000000000004fXFwHzGuTlkQcR5z1rzEXbWAnrZUJOq5Ua3YnFcvvx",
        "ids": "<1 hidden ids>",
        "y": 88,
        "x": 123123
    }
    >>> ldict(y=88, x=123123) != ldict(x=88, y=123123)  # Ids of original values embbed the keys.
    True
    >>> d = ldict(x=123123, y=88)
    >>> e = d >> (lambda x: {"z": x**2}) >> (lambda x,y: {"w": x/y})
    >>> e.show(colored=False)
    {
        "id": "yk81NHFkhUQg.3W8CiZjeayh8367tNNSrQu9rVtqkCDHcaaBiju6GMVtyxp.h1Ac",
        "ids": {
            "w": "94Dfzc9y5U0BIYL6NeE-oidtGaA28wYqyASJAV39umabOWr9Iz0PseJTunWMwA8G",
            "z": "pfwOevvOc0THi7b1R4kjIhqtqUgM.-hm1DsEfbPnE4cYsXnffFhkXvSDXzt6mQVV",
            "x": "0000000000000000000001rvYECRwLyX0-V2zUk5HJ.QE5wqO8-0OdDQBV4vkwcE",
            "y": "0000000000000000000002QrIU4K9LkrjRjOxGJlTLXniuT1bLMpZ8glu3j9Te0f"
        },
        "w": "<unevaluated lazy field>",
        "z": "<unevaluated lazy field>",
        "x": 123123,
        "y": 88
    }
    >>> a = d >> (lambda x: {"z": x**2}) >> (lambda x,y: {"w": x/y})
    >>> b = d >> (lambda x,y: {"w": x/y}) >> (lambda x: {"z": x**2})
    >>> a != b  # Calculated values are order-sensitive.
    True
    """

    def __init__(self, /, _dictionary=None, keepblob=False, version="UT64.4", **kwargs) -> None:
        super().__init__()
        if version == "UT32.4":
            self.identity = identity32
            self.rho0 = Hosh.fromid("------------------------------.0", version="UT32.4")
        elif version == "UT64.4":
            self.identity = identity64
            self.rho0 = Hosh.fromid(
                "--------------------------------------------------------------.0", version="UT64.4"
            )
        else:
            raise Exception("Unknown version:", version)

        self.keepblob, self.hosh, self.version = keepblob, self.identity, version
        self.hoshes, self.blobs, self.previous = {}, {}, {}
        self.data: Dict[str, VT] = {"id": self.hosh.id, "ids": {}}  # 'id' will always be the first field
        if _dictionary is not None:
            self.update(_dictionary)  # REMINDER: update() acts on self.data.
        if kwargs:
            self.update(kwargs)

    def __getitem__(self, field: str) -> VT:
        if field in self.data:
            content = self.data[field]
            if callable(content):
                if field in self.previous:
                    self.data[field] = self.previous[field].pop()
                    if len(self.previous[field]) == 0:
                        del self.previous[field]
                self.data[field] = content(**self._getargs(field, content))
                return self.data[field]
            else:
                return self.data[field]
        raise KeyError(field)

    def _getargs(self, field, f, fargs=None):
        fargs = fargs or signature(f).parameters.keys()
        dic = {}
        for k in fargs:
            if k not in self:
                print(self)
                raise Exception(f"Missing field {k} needed by {f} to calculate field {field}.")
            dic[k] = self[k]
        return dic

    def _trigger(self, field, f, fargs):
        def closure():
            dic = f(**self._getargs(field, f, fargs))
            for arg, value in dic.items():
                self.data[arg] = value
            return self[field]

        return closure

    def __setitem__(self, field: str, value: VT) -> None:
        if not isinstance(field, str):
            raise Exception(f"Key must be string, not {type(field)}.", field)

        # Work around field overwrite.
        if field in self.data:
            if callable(value):
                if field in self.previous:
                    self.previous[field].append(self.data[field])
                else:
                    self.previous[field] = [self.data[field]]
                # TODO: histórico aqui vai ser aumentado em: f
            else:
                del self[field]
                # TODO: histórico aqui vai ser aumentado em: (---...----x)x'

        # Create field's hosh and update ldict's hosh.
        h, blob = process(field, value, version=self.version)
        old_hash = self.hosh
        self.hosh *= h
        field_hash = self.hosh / old_hash if blob is None else h
        self.hoshes[field] = field_hash

        # Update data.
        self.data[field] = value
        self.data["id"] = self.hosh.id
        self.data["ids"][field] = field_hash.id

        # Keep blob if required. TODO: keep hosh sem key embutida
        if blob and self.keepblob:
            self.blobs[field] = blob

    def __delitem__(self, field: str) -> None:
        # REMINDER: esse del não cria placeholder, para ficar intuitivo manipular um dict. põe x, tira x etc.
        #           é distinto da remoção by name/index.
        del self.data[field]
        del self.hoshes[field]
        del self.data["ids"][field]
        self.hosh = self.identity
        for hosh in self.hoshes.values():
            self.hosh *= hosh
        # TODO: histórico aqui vai ser p/ del d[y] : [yzw]-¹zw

    def __rshift__(self, f: Union[Dict, Callable]):
        """Used for multiple return values, or to insert values during a multistep process.
        >>> from ldict import ø
        >>> d = ø >> {"x": 1} >> (lambda x: {"y": x**2}) >> (lambda x,y:{"z": x+y, "w": x/y})
        >>> d.show(colored=False)
        {
            "id": "QNRxg0I5naPZX-dls7VQ622Htt-Q1Yfx2IL1DiJH0EtihD-yzK8y1qb7KysmVigM",
            "ids": {
                "z": "scVKB6ugOhLvjEhlIt4JPhWJl1qTZB1Y4SbQIOy55UbEm2bEumU8bfRktf0ZDWCc",
                "w": "111111110GZ11111114bRi9Np1yhkc31s9Trh6HU.4cocBk71K00FEal5z-gv8Db",
                "y": "nzWNFVcPAeftDkW-KFMU9VXfh1B1uyzVMcIhuAtpHkIsu4i3VaYOlUd8hgCK7TjI",
                "x": "0000000000000000000004nMDSf.TnQ5bqaZyLaOm1jkIt.q0Kxrj1aVGPmDOeqb"
            },
            "z": "<unevaluated lazy field>",
            "w": "<unevaluated lazy field>",
            "y": "<unevaluated lazy field>",
            "x": 1
        }
        """
        # TODO: d >> {"x": None}   delete by name
        #   histórico vai ser aumentado em : (---...----x)    e vai placeholder em d
        # TODO: d >> {2: None}   delete by index
        clone = self.copy()
        if isinstance(f, dict):
            for k, v in f.items():
                clone[k] = v
            return clone
        elif not callable(f):
            raise Exception("f should be callable or dict.")

        # Extract output fields. https://stackoverflow.com/a/68753149/9681577
        out = StringIO()
        decompile(bytecode_version=None, co=f.__code__, out=out)
        ret = "".join([line for line in out.getvalue().split("\n") if not line.startswith("#")])
        if "return" not in ret:
            print(ret)
            raise Exception(f"Missing return statement:")
        dicts = re.findall("(?={)(.+?)(?<=})", ret)
        if len(dicts) != 1:
            raise Exception(
                "Cannot detect output fields:" "Missing dict (with pairs 'identifier'->result) as a return value.",
                dicts,
                ret,
            )
        output_fields = re.findall(r"(?:[\"'])([a-zA-Z]+[a-zA-Z0-9_]*)(?:[\"'])", dicts[0])

        # Work around field overwrite.
        for field in output_fields:
            if field in clone.data:
                clone.previous[field] = clone.data[field]

        # Detect input fields.
        fargs = signature(f).parameters.keys()

        # Attach hosh to f if needed.
        if not hasattr(f, "hosh"):
            f.hosh = fhosh(f, version=self.version)

        # Update clone id.
        if f.hosh.etype != "ordered":
            raise OverwriteException("Function element probably will never be allowed to have etype!=ordered.")
        old_hash = clone.hosh
        clone.hosh *= f.hosh
        clone.data["id"] = clone.id
        ufu = clone.hosh * ~old_hash  # z = ufu-¹

        # Add triggers for future evaluation.
        acc = self.identity
        new_hashes = {}
        new_data = {"id": clone.id, "ids": {}}
        for i, field in enumerate(output_fields):
            if len(output_fields) == 1:
                field_hash = ufu
            else:
                if i < len(output_fields) - 1:
                    field_hash = ufu * (self.rho0 + Hosh.fromn(i, self.version))
                    acc *= field_hash
                else:
                    field_hash = ~acc * ufu
            new_hashes[field] = field_hash
            # noinspection PyProtectedMember
            new_data[field] = clone._trigger(field, f, fargs)
            new_data["ids"][field] = field_hash.id

        new_hashes.update(clone.hoshes)
        new_data["ids"].update(clone.data["ids"])
        del clone.data["ids"]
        new_data.update(clone.data)

        clone.hoshes = new_hashes
        clone.data = new_data
        return clone


class OverwriteException(Exception):
    pass



import dis

from garoupa import Hosh
from orjson import dumps, OPT_SORT_KEYS


def process(field, value, version):
    """
    Create hosh with etype=
        "hybrid" if 'value' is not callable
        "ordered" if 'value' is a callable without an attribute 'hosh'

    Usage:

    >>> hosh, blob = process("X", 123, "UT64.4")
    >>> hosh.id
    '0000000000000000000006jfeYEWNkSX84ldc-BIj0ST9rhlkTESf81XaQdiwRPm'
    >>> blob
    b'{"X":123}'
    >>> f = lambda: 123
    >>> hasattr(f, "hosh")
    False
    >>> hosh, blob = process("X", f, "UT64.4")
    >>> hosh.id
    'jq-fkk.Ejr0v.hOGHyBl4HaicbeEu-34uuJ8wQQNCZVqWy-YDE2Ser6JkaFjd-iH'
    >>> blob is None
    True
    >>> f.hosh = hosh
    >>> hasattr(f, "hosh")
    True
    >>> hosh2, blob = process("X", f, "UT64.4")
    >>> hosh is hosh2
    True

    Parameters
    ----------
    field
    value
    version

    Returns
    -------

    """
    if callable(value):
        h = value.hosh if hasattr(value, "hosh") else fhosh(value, version)
        return h, None
    obj = {field: value}
    # TODO: separar key do hosh pra ser usada como no artigo, para localizar o blob no db
    bytes = dumps(obj, option=OPT_SORT_KEYS)
    return Hosh(bytes, "hybrid", version=version), bytes


def fhosh(f, version):
    """
    Create hosh with etype="ordered" using bytecodeof "f" as binary content.

    Usage:

    >>> print(fhosh(lambda x: {"z": x**2}, "UT64.4"))
    pfwOevvOc0THi7b1R4kjIm.dfJ-YO6FPn6o5YhsOOoVNLHy4dMosf2aTkAQW5Us7

    Parameters
    ----------
    f
    version

    Returns
    -------

    """
    # Clean line numbers.
    groups = [l for l in dis.Bytecode(f).dis().split("\n\n") if l]
    clean = []
    for group in groups:
        lines = [segment for segment in group.split(" ") if segment][1:]
        clean.append(lines)

    return Hosh(dumps(clean), "ordered", version=version)



class Empty:
    def __rshift__(self, other):
        """
        Usage:

        >>> from ldict import ø
        >>> d = ø >> {"x": 2}
        >>> d.show(colored=False)
        {
            "id": "000000000000000000000c3aop1df5AZXCRMY3yInQeUYccGQRclWo8TvfKPB4YT",
            "ids": {
                "x": "000000000000000000000c3aop1df5AZXCRMY3yInQeUYccGQRclWo8TvfKPB4YT"
            },
            "x": 2
        }
        """
        from ldict import Ldict
        return Ldict(other)



import json
import re
from typing import Iterator, Iterable, TypeVar, Union
from typing import TYPE_CHECKING

from garoupa import Hosh

if TYPE_CHECKING:
    from ldict import Ldict

VT = TypeVar("VT")


class Aux:
    hosh: Hosh
    hoshes: dict
    data: dict
    keepblob: bool
    version: str

    @property
    def n(self):
        """
        Usage:

        >>> from ldict import ldict
        >>> ldict(x=134124).n
        78297399190853865925546422718430264658855010523272162045457933175107322571720
        """
        return self.hosh.n

    @property
    def id(self):
        """
        Usage:

        >>> from ldict import ldict
        >>> ldict(x=134124).id
        '000000000000000000000aQqMuGtzigXKY9vs4vME6FPSHKoBxSJXSG4.IDSRKL8'
        >>> ldict(x=134124).ids
        {'x': '000000000000000000000aQqMuGtzigXKY9vs4vME6FPSHKoBxSJXSG4.IDSRKL8'}
        """
        return self.hosh.id

    @property
    def idc(self):
        """Colored id"""
        return self.hosh.idc

    @property
    def ids(self):
        return self.data["ids"]

    @property
    def all(self):
        """
        Usage:

        >>> from ldict import ldict
        >>> out = ldict(x=134124, y= 56).all
        >>> print(ansi_escape.sub('', out))  # doctest: +SKIP
        {
            "id": "LI2jBaxgZwh8khoYlfAYikEI8gnGnN5NX2h6kBwL6v2",
            "x": 134124,
            "id_x": "MO72GzebQLg1Q6EfBqPlpor9I5P7XXDByDrXsj9kdSS",
            "y": 56,
            "id_y": "xbfZfXwFFza8UwbSM6HM5ig8VqYmwXewLkX7LW7SAV6"
        }
        """
        return self.__repr__(all=True)

    def show(self, colored=True):
        """
        Usage:

        >>> from ldict import ldict
        >>> ldict(x=134124, y= 56).show(colored=False)
        {
            "id": "000000000000000000000cBXR5qArL1WVYOn7MW0MWyWhPWP31DmatnNd5soVRK0",
            "ids": {
                "x": "000000000000000000000aQqMuGtzigXKY9vs4vME6FPSHKoBxSJXSG4.IDSRKL8",
                "y": "0000000000000000000001Nx4CM6UsN0b0ETHIqg8ER6r8cqtvNh1mJIdoQy3kRR"
            },
            "x": 134124,
            "y": 56
        }
        """
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        return print(self.all if colored else ansi_escape.sub("", self.all))

    @property
    def cells(self):
        """
        For new untransformed data, it is always the identity permutation.
        Usage:

        >>> from ldict import ldict
        >>> a = ldict(x=134124)
        >>> b = ldict(y=134124)
        >>> a.cells
        [0, 0, 12473495331365790899, 7728865973021107401, 8212075338224970119, 6157470389993991581]
        >>> (a + b).cells
        [0, 0, 16031506638198821588, 13834746703073065004, 4696393870013716036, 6637745325376246119]
        >>> b = a + ldict(y=134124)
        >>> b.cells
        [0, 0, 16031506638198821588, 13834746703073065004, 4696393870013716036, 6637745325376246119]
        >>> f = lambda x: {'z': x ** 2}
        >>> print(b >> f)
        {
            "id": "pfwOevvOc0THi7b1R4kjIeBw1RzBLiLT3uxe.T0Cob2a9xr56v1mMJIIvBjUkNKm",
            "ids": "<2 hidden ids>",
            "z": "<unevaluated lazy field>",
            "x": 134124,
            "y": 134124
        }
        """
        return self.hosh.cells

    def __str__(self, all=False):
        dic = self.data.copy()
        for k, v in self.data.items():
            if callable(v):
                dic[k] = "<unevaluated lazy field>"
        if not all:
            dic["ids"] = "<1 hidden id>" if len(dic["ids"]) == 1 else f"<{(len(self) - 1) // 2} hidden ids>"
        return json.dumps(dic, indent=4)

    def __repr__(self, all=False):
        dic = self.data.copy()
        for k, v in self.data.items():
            if callable(v):
                dic[k] = "<unevaluated lazy field>"
        if not all:
            dic["ids"] = "<1 hidden id>" if len(dic["ids"]) == 1 else f"<{(len(self) - 1) // 2} hidden ids>"
        txt = json.dumps(dic, indent=4)
        for k, v in dic.items():
            if k == "id":
                txt = txt.replace(dic[k], self.hosh.idc)
        if all:
            for k, v in self.hoshes.items():
                txt = txt.replace(v.id, v.idc)  # REMINDER: workaround to avoid json messing with colors
        return txt

    def __add__(self, other):
        """
        Usage:

        >>> from ldict import ldict
        >>> a = ldict(x=134124)
        >>> b = ldict(y=542542)
        >>> a.id, b.id
        ('000000000000000000000aQqMuGtzigXKY9vs4vME6FPSHKoBxSJXSG4.IDSRKL8', '0000000000000000000008RYrX8XDiNbFnEv9cJZy-pTw2nSHxy64stPR602KmI1')
        >>> print(a + b)
        {
            "id": "0000000000000000000003GncpPpaB5PojN-BhdJSHHHmK6fh4aZ3j7UQODURT5o",
            "ids": "<1 hidden ids>",
            "x": 134124,
            "y": 542542
        }
        >>> print(b + a)
        {
            "id": "0000000000000000000003GncpPpaB5PojN-BhdJSHHHmK6fh4aZ3j7UQODURT5o",
            "ids": "<1 hidden ids>",
            "y": 542542,
            "x": 134124
        }
        >>> print(b + a + b)
        {
            "id": "0000000000000000000003GncpPpaB5PojN-BhdJSHHHmK6fh4aZ3j7UQODURT5o",
            "ids": "<1 hidden ids>",
            "x": 134124,
            "y": 542542
        }
        """

        from ldict import ldict

        new = ldict(self)
        new.update(other)
        return new

    def copy(self):
        from ldict import ldict

        obj = ldict()
        # mutability WARNING: any attempt to inplace update nested structures will be disastrous, e.g.: d.d["x"] = 5
        # TODO: solution: freeze ldict when inserting, so that it cannot mutate anymore (deletion/insertion/update)
        #   It can be a wrapper class around _set_ and _del_
        obj.data = self.data.copy()
        obj.hoshes = self.hoshes.copy()
        obj.hosh = self.hosh
        obj.previous = self.previous.copy()
        obj.blobs = self.blobs.copy()
        obj.keepblob = self.keepblob
        obj.version = self.version
        return obj

    @property
    def hex(self):
        return hex(self.n)[2:]

    def __hash__(self):
        return self.n

    def __eq__(self, other):
        from ldict import Ldict

        if isinstance(other, Ldict):
            return self.n == other.n
        return NotImplemented

    def __ne__(self, other):
        return self.hosh != other.hosh

    def __contains__(self, field: str) -> bool:
        return field in self.data

    def __getattr__(self, item):
        if item in self:
            return self[item]
        return self.__getattribute__(item)

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    @classmethod
    def fromkeys(cls, iterable: Iterable[str], value: VT) -> "Union[Ldict, Aux]":
        """Create a new dictionary with keys from `iterable` and values set to `value`.

        Args:
            iterable: A collection of keys.
            value: The default value. It defaults to None.
            All of the values refer to just a single instance,
            so it generally doesn’t make sense for value to be a mutable object such as an empty list.
            To get distinct values, use a dict comprehension instead.

        Returns:
            A new instance of Dict.
        """
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def keys(self):
        yield from self.data.keys()

    def update(self, other=(), /, **kwds):
        """Update the dictionary with the key/value pairs from other, overwriting existing keys.
        Usage:

        >>> from ldict import ldict
        >>> a = ldict(x=123)
        >>> print(a)
        {
            "id": "000000000000000000000eqEFMZGUZzQFTmxCZxEnDEJHiWcdZc1I4.oNUufZzv9",
            "ids": "<1 hidden id>",
            "x": 123
        }
        >>> b = ldict(y="some text")
        >>> print(b)
        {
            "id": "0000000000000000000004uDaFL.-o9CHFrbEUsuhPnJHDTFe5Uoz746KCdo24Om",
            "ids": "<1 hidden id>",
            "y": "some text"
        }
        >>> b.update(a)
        >>> print(b)
        {
            "id": "0000000000000000000002VfQqJGTlN7lwNJfR-6l1ErmWNRs3SzhY3vwuHDhpYF",
            "ids": "<1 hidden ids>",
            "y": "some text",
            "x": 123
        }
        """
        from ldict import Ldict

        if isinstance(other, Ldict):
            if self.version != other.version:
                raise Exception("Different versions:", self.version, other.version)
            self.keepblob = self.keepblob or other.keepblob
            self.blobs.update(other.blobs)
        kwds.update(other)
        for field, value in kwds.items():
            if field not in ["id", "ids"]:
                self[field] = value



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ldict', 'ldict.abs', 'ldict.abs.mixin']

package_data = \
{'': ['*']}

install_requires = \
['garoupa>=2.210811.7,<3.0.0',
 'orjson>=3.5.0,<4.0.0',
 'uncompyle6>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'ldict',
    'version': '2.210815.2',
    'description': 'Lazy dict with universaly unique identified values',
    'long_description': '![test](https://github.com/davips/ldict/workflows/test/badge.svg)\n[![codecov](https://codecov.io/gh/davips/ldict/branch/main/graph/badge.svg)](https://codecov.io/gh/davips/ldict)\n\n# ldict\nUniquely identified lazy dict.\n\n[Latest version](https://github.com/davips/ldict)\n\n## Installation\n### as a standalone lib.\n```bash\n# Set up a virtualenv. \npython3 -m venv venv\nsource venv/bin/activate\n\n# Install from PyPI...\npip install --upgrade pip\npip install -U ldict\n\n# ...or, install from updated source code.\npip install git+https://github.com/davips/ldict\n```\n\n### from source\n```bash\ngit clone https://github.com/davips/ldict\ncd ldict\npoetry install\n```\n\n## Examples\n**Merging two ldicts**\n<details>\n<p>\n\n```python3\nfrom ldict import ldict\n\na = ldict(x=3)\nprint(a)\n"""\n{\n    "id": "000000000000000000000dCguKC6bS.95Bqe35gNJir04CxRv6dY1dPRO7R3PTKa",\n    "ids": "<1 hidden id>",\n    "x": 3\n}\n"""\n```\n\n```python3\n\nb = ldict(y=5)\nprint(b)\n"""\n{\n    "id": "000000000000000000000dlrfIktAZFYqwNrmOdFDZKV966TGLM7hu0fllFfC8Kt",\n    "ids": "<1 hidden id>",\n    "y": 5\n}\n"""\n```\n\n```python3\n\nprint(a + b)\n"""\n{\n    "id": "000000000000000000000aXHKqWzMQIMw6bFpTur11RVdIEJ9SLzyXQ57tuiIAfV",\n    "ids": "<1 hidden ids>",\n    "x": 3,\n    "y": 5\n}\n"""\n```\n\n\n</p>\n</details>\n\n## Features (current or planned)\n* [x] \n* [ ] \n\n## How to use [outdated]\nTwo main entities are identifiable: processing functions and bags of values.\nA processing function is any callable the follows the rules below.\nA bag of values is a ldict object. It is a mapping between string keys, called fields,\nand any serializable object.\nThe bag id (identifier) and the field ids are also part of the mapping.  \n\nThe user should provide a unique identifier for each function object.\nIt should be put as a 43 digits long base-62 string under the key "_id", or, \nalternatively, a Hosh object inside the returned dict, under the key "_id".\nThe only exception is when using the assignment syntax, \nbecause the return value is the proper result of the calculation.\nWhen using the assignment syntax, it is assumed the \'id\' should be automatically \ncalculated by the bytecode obtained through source code parsing.\nFor this reason, such functions should be simple, i.e., \nwith minimal external dependencies, to avoid the unfortunate situation where two\nfunctions with identical local code actually perform different calculations through\ncalls to external code that implement different algorithms with the same name.\n\nOne way to emulate such behavior for the function application syntax (a >> f) is to\nexplicitly refuse to provide a hash/id.\nThis can be done by setting `"_id": None` inside the returned dictionary.\n',
    'author': 'davips',
    'author_email': 'dpsabc@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
